@@ -1678,9 +1678,20 @@ frame_flags(Frame) -->
     },
     [ '~w~w '-[T, S] ].
 
-% trace/1,2 context handling
+% trace/1 context handling
+port(Port, Dict) -->
+    { _{level:Level, start:Time} :< Dict
+    },
+    (   { Port \== call,
+          get_time(Now),
+          Passed is (Now - Time)*1000.0
+        }
+    ->  [ '[~d +~1fms] '-[Level, Passed] ]
+    ;   [ '[~d] '-[Level] ]
+    ),
+    port(Port).
 port(Port, _Id-Level) -->
-    [ '[~d] '-Level ],
+    [ '[~d] '-[Level] ],
     port(Port).
 
 port(Port) -->
@@ -1930,7 +1941,7 @@ print_message(Level, Term) :-
     ;   true
     ).
 
-push_msg(Term, Stack) :-
+push_msg(Term, Messages) :-
     nb_current('$inprint_message', Messages),
     !,
     \+ ( '$member'(Msg, Messages),
@@ -1938,7 +1949,7 @@ push_msg(Term, Stack) :-
        ),
     Stack = [Term|Messages],
     b_setval('$inprint_message', Stack).
-push_msg(Term, [Term]) :-
+push_msg(Term, []) :-
     b_setval('$inprint_message', [Term]).
 
 pop_msg(Stack) :-
