@@ -3,7 +3,8 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2019, VU University Amsterdam
+    Copyright (c)  2019-2020, VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -37,6 +38,7 @@
             trimcore/0,
 
             abolish_table_info/0,
+            close_open_tables/1,          % ?
 
             str_cat/3,
 
@@ -80,6 +82,13 @@ trimcore :-
 
 abolish_table_info.
 
+%!  close_open_tables(?Arg)
+%
+%   Undocumented in the XSB manual. Tables are always closed on
+%   exceptions, so it is unclear what this should do?
+
+close_open_tables(_).
+
 %!  str_cat(+Atom1, +Atom2, -Atom3)
 
 str_cat(A, B, AB) :-
@@ -99,12 +108,12 @@ parsort(_List, _Spec, Dupl, _Sorted) :-
     var(Dupl),
     !,
     uninstantiation_error(Dupl).
-parsort(List, asc,  0, Sorted) :- !, sort(0, @<,  List, Sorted).
-parsort(List, asc,  _, Sorted) :- !, sort(0, @=<, List, Sorted).
-parsort(List, [],   0, Sorted) :- !, sort(0, @<,  List, Sorted).
-parsort(List, [],   _, Sorted) :- !, sort(0, @=<, List, Sorted).
-parsort(List, desc, 0, Sorted) :- !, sort(0, @>,  List, Sorted).
-parsort(List, desc, _, Sorted) :- !, sort(0, @>=, List, Sorted).
+parsort(List, asc,  0, Sorted) :- !, sort(0, @=<, List, Sorted).
+parsort(List, asc,  _, Sorted) :- !, sort(0, @<,  List, Sorted).
+parsort(List, [],   0, Sorted) :- !, sort(0, @=<, List, Sorted).
+parsort(List, [],   _, Sorted) :- !, sort(0, @<,  List, Sorted).
+parsort(List, desc, 0, Sorted) :- !, sort(0, @>=, List, Sorted).
+parsort(List, desc, _, Sorted) :- !, sort(0, @>,  List, Sorted).
 parsort(List, SortSpec, Dupl, Sorted) :-
     must_be(list, SortSpec),
     reverse(SortSpec, Rev),
@@ -115,10 +124,10 @@ parsort_([H|T], Dupl, List0, List) :-
     parsort_1(H, Dupl, List0, List1),
     parsort_(T, Dupl, List1, List).
 
-parsort_1(asc(I),  0, List, Sorted) :- !, sort(I, @<,  List, Sorted).
-parsort_1(asc(I),  _, List, Sorted) :- !, sort(I, @=<, List, Sorted).
-parsort_1(desc(I), 0, List, Sorted) :- !, sort(I, @>,  List, Sorted).
-parsort_1(desc(I), _, List, Sorted) :- !, sort(I, @>=, List, Sorted).
+parsort_1(asc(I),  0, List, Sorted) :- !, sort(I, @=<, List, Sorted).
+parsort_1(asc(I),  _, List, Sorted) :- !, sort(I, @<,  List, Sorted).
+parsort_1(desc(I), 0, List, Sorted) :- !, sort(I, @>=, List, Sorted).
+parsort_1(desc(I), _, List, Sorted) :- !, sort(I, @>,  List, Sorted).
 parsort_1(Spec,  _, _, _) :-
     domain_error(parsort_spec, Spec).
 
