@@ -45,15 +45,17 @@
 	    load_test_files/1,          % +Options
 	    running_tests/0,            % Prints currently running test
 	    current_test/5,             % ?Unit,?Test,?Line,?Body,?Options
+	    current_test_unit/2,        % ?Unit,?Options
 	    test_report/1               % +What
 	  ]).
 
 /** <module> Unit Testing
 
 Unit testing environment for SWI-Prolog and   SICStus Prolog. For usage,
-please visit http://www.swi-prolog.org/pldoc/package/plunit.
+please visit https://www.swi-prolog.org/pldoc/package/plunit.
 */
 
+:- autoload(library(statistics), [call_time/2]).
 :- autoload(library(apply), [maplist/3,include/3]).
 :- autoload(library(lists), [member/2,append/2]).
 :- autoload(library(option), [option/3,option/2]).
@@ -62,6 +64,8 @@ please visit http://www.swi-prolog.org/pldoc/package/plunit.
 :- autoload(library(error), [must_be/2]).
 :- autoload(library(thread), [concurrent_forall/2]).
 :- autoload(library(aggregate), [aggregate_all/3]).
+:- autoload(library(streams), [with_output_to/3]).
+:- autoload(library(ansi_term), [ansi_format/3]).
 
 :- meta_predicate valid_options(+, 1).
 
@@ -192,44 +196,49 @@ user:term_expansion((:- thread_local(PI)), (:- dynamic(PI))) :-
    (   current_test_flag(test_options, _)
    ->  true
    ;   set_test_flag(test_options,
-		 [ run(make),       % run tests on make/0
-		   sto(false)
-		 ])
+                     [ run(make),       % run tests on make/0
+                       sto(false),
+                       output(on_failure)
+                     ])
    ).
 
 %!  set_test_options(+Options)
 %
 %   Specifies how to deal with test suites.  Defined options are:
 %
-%           * load(+Load)
-%           Whether or not the tests must be loaded.  Values are
-%           =never=, =always=, =normal= (only if not optimised)
+%    - load(+Load)
+%      Whether or not the tests must be loaded.  Values are
+%      `never`, `always`, `normal` (only if not optimised)
 %
-%           * run(+When)
-%           When the tests are run.  Values are =manual=, =make=
-%           or make(all).
+%    - run(+When)
+%      When the tests are run.  Values are `manual`, `make`
+%      or make(all).
 %
-%           * silent(+Bool)
-%           If =true= (default =false=), report successful tests
-%           using message level =silent=, only printing errors and
-%           warnings.
+%    - silent(+Bool)
+%      If `true` (default `false`), report successful tests
+%      using message level `silent`, only printing errors and
+%      warnings.
 %
-%           * sto(+Bool)
-%           How to test whether code is subject to occurs check
-%           (STO).  If =false= (default), STO is not considered.
-%           If =true= and supported by the hosting Prolog, code
-%           is run in all supported unification mode and reported
-%           if the results are inconsistent.
+%    - output(+When)
+%      If `always`, emit all output as it is produced, if `never`,
+%      suppress all output and if `on_failure`, emit the output
+%      if the test fails.
 %
-%           * cleanup(+Bool)
-%           If =true= (default =false), cleanup report at the end
-%           of run_tests/1.  Used to improve cooperation with
-%           memory debuggers such as dmalloc.
+%    - sto(+Bool)
+%      How to test whether code is subject to occurs check
+%      (STO).  If `false` (default), STO is not considered.
+%      If `true` and supported by the hosting Prolog, code
+%      is run in all supported unification mode and reported
+%      if the results are inconsistent.
 %
-%           * concurrent(+Bool)
-%           If =true= (default =false), run all tests in a block
-%           concurrently.
+%    - cleanup(+Bool)
+%      If `true` (default =false), cleanup report at the end
+%      of run_tests/1.  Used to improve cooperation with
+%      memory debuggers such as dmalloc.
 %
+%    - concurrent(+Bool)
+%      If `true` (default `false`), run all tests in a unit
+%      concurrently.
 
 set_test_options(Options) :-
     valid_options(Options, global_test_option),
@@ -237,6 +246,8 @@ set_test_options(Options) :-
 
 global_test_option(load(Load)) :-
     must_be(oneof([never,always,normal]), Load).
+global_test_option(output(Cond)) :-
+    must_be(oneof([always,never, on_failure]), Cond).
 global_test_option(run(When)) :-
     must_be(oneof([manual,make,make(all)]), When).
 global_test_option(silent(Bool)) :-
@@ -562,14 +573,42 @@ test_set_option(sto(V)) :-
 test_set_option(concurrent(V)) :-
     must_be(boolean, V).
 
+		 /*******************************
+		 *             UTIL		*
+		 *******************************/
+
+:- meta_predicate
+       reify(0, -),
+       capture_output(0,-,+).
+
+%!  reify(:Goal, -Result) is det.
+
+reify(Goal, Result) :-
+    (   catch(Goal, E, true)
+    ->  (   var(E)
+	->  Result = true
+	;   Result = throw(E)
+	)
+    ;   Result = false
+    ).
+
+capture_output(Goal, Output, Options) :-
+    option(output(How), Options, always),
+    (   How == always
+    ->  call(Goal)
+    ;   with_output_to(string(Output), Goal,
+                       [ capture([user_output, user_error])])
+    ).
+
 
 		 /*******************************
 		 *        RUNNING TOPLEVEL      *
 		 *******************************/
 
 :- thread_local
+    test_count/1,                   % Count
     passed/5,                       % Unit, Test, Line, Det, Time
-    failed/4,                       % Unit, Test, Line, Reason
+    failed/5,                       % Unit, Test, Line, Reason, Time
     failed_assertion/7,             % Unit, Test, Line, ALoc, STO, Reason, Goal
     blocked/4,                      % Unit, Test, Line, Reason
     sto/4,                          % Unit, Test, Line, Results
@@ -590,29 +629,23 @@ test_set_option(concurrent(V)) :-
 %   within a unit.
 
 run_tests :-
+    findall(Unit, current_test_unit(Unit,_), Units),
+    run_tests(Units).
+
+run_tests(Set) :-
     cleanup,
+    count_tests(Set, Count),
+    asserta(test_count(Count)),
     setup_call_cleanup(
 	setup_trap_assertions(Ref),
-	run_current_units,
+	run_unit_and_check_errors(Set),
 	report_and_cleanup(Ref)).
 
-run_current_units :-
-    forall(current_test_set(Set),
-	   run_unit(Set)),
-    all_tests_passed(_).
-
 report_and_cleanup(Ref) :-
     cleanup_trap_assertions(Ref),
     report,
     cleanup_after_test.
 
-run_tests(Set) :-
-    cleanup,
-    setup_call_cleanup(
-	setup_trap_assertions(Ref),
-	run_unit_and_check_errors(Set),
-	report_and_cleanup(Ref)).
-
 run_unit_and_check_errors(Set) :-
     run_unit(Set),
     all_tests_passed(_).
@@ -626,22 +659,50 @@ run_unit(Spec) :-
     unit_from_spec(Spec, Unit, Tests, Module, UnitOptions),
     (   option(blocked(Reason), UnitOptions)
     ->  info(plunit(blocked(unit(Unit, Reason))))
-    ;   setup(Module, unit(Unit), UnitOptions)
-    ->  get_time(T0),
-	info(plunit(begin(Spec))),
-	run_unit_2(Unit, Tests, Module, UnitOptions),
-	get_time(T1),
-	test_summary(Unit, Summary),
-	Time is T1-T0,
-	info(plunit(end(Spec, Summary.put(time, Time)))),
-	(   message_level(silent)
-	->  true
-	;   format(user_error, '~N', [])
-	),
-	cleanup(Module, UnitOptions)
+    ;   condition(Module, unit(Unit), UnitOptions)
+    ->  (   setup(Module, unit(Unit), UnitOptions)
+        ->  info(plunit(begin(Spec))),
+            call_time(run_unit_2(Unit, Tests, Module, UnitOptions), Time),
+            test_summary(Unit, Summary),
+            info(plunit(end(Spec, Summary.put(time, Time)))),
+            cleanup(Module, UnitOptions)
+        )
     ;   true
     ).
 
+%!  count_tests(+Spec, -Count) is det.
+%
+%   Count the number of tests to run.
+
+count_tests(Spec, Count) :-
+    count_tests(Spec, 0, Count).
+
+count_tests([], Count, Count) :-
+    !.
+count_tests([H|T], Count0, Count) :-
+    !,
+    count_tests(H, Count0, Count1),
+    count_tests(T, Count1, Count).
+count_tests(Spec, Count0, Count) :-
+    unit_from_spec(Spec, Unit, Tests, _Module, UnitOptions),
+    (   option(blocked(_Reason), UnitOptions)
+    ->  Count = Count0
+    ;   var(Tests)
+    ->  count(( current_test(Unit,_,_,_,TestOptions),
+                \+ option(blocked(_), TestOptions)), N),
+	Count is Count0+N
+    ;   atom(Tests),
+	current_test(Unit,Tests,_,_,_)
+    ->  Count is Count0+1
+    ;   is_list(Tests)
+    ->  count((member(T, Tests), current_test(Unit,T,_,_,_)), N),
+	Count is Count0+N
+    ;   Count = Count0
+    ).
+
+
+
+
 :- if(current_prolog_flag(threads, true)).
 run_unit_2(Unit, Tests, Module, UnitOptions) :-
     option(concurrent(true), UnitOptions, false),
@@ -681,8 +742,10 @@ matching_test(Name, Set) :-
 
 cleanup :-
     thread_self(Me),
+    set_flag(plunit_test, 1),
+    retractall(test_count(_)),
     retractall(passed(_, _, _, _, _)),
-    retractall(failed(_, _, _, _)),
+    retractall(failed(_, _, _, _, _)),
     retractall(failed_assertion(_, _, _, _, _, _, _)),
     retractall(blocked(_, _, _, _)),
     retractall(sto(_, _, _, _)),
@@ -850,9 +913,10 @@ run_test_once(Unit, Name, Line, Options, Body) :-
     !,
     current_unification_capability(Type),
     begin_test(Unit, Name, Line, Type),
-    run_test_6(Unit, Name, Line, Options, Body, Result),
+    capture_output(run_test_6(Unit, Name, Line, Options, Body, Result),
+                   Output, GlobalOptions),
     end_test(Unit, Name, Line, Type),
-    report_result(Result, Options).
+    report_result(Result, Output, Options).
 run_test_once(Unit, Name, Line, Options, Body) :-
     current_unit(Unit, _Module, _Supers, UnitOptions),
     option(sto(Type), UnitOptions),
@@ -867,24 +931,29 @@ run_test_once(Unit, Name, Line, Options, Body) :-
 		 set_unification_capability(Cap0)).
 
 run_test_cap(Unit, Name, Line, Options, Body) :-
+    current_test_flag(test_options, GlobalOptions),
     (   option(sto(Type), Options)
     ->  unification_capability(Type),
 	set_unification_capability(Type),
 	begin_test(Unit, Name, Line, Type),
-	run_test_6(Unit, Name, Line, Options, Body, Result),
+	capture_output(run_test_6(Unit, Name, Line, Options, Body, Result),
+                       Output, GlobalOptions),
 	end_test(Unit, Name, Line, Type),
-	report_result(Result, Options)
-    ;   findall(Key-(Type+Result),
-		test_caps(Type, Unit, Name, Line, Options, Body, Result, Key),
-		Pairs),
+	report_result(Result, Output, Options)
+    ;   findall(Key-(r(Type,Result,Output)),
+                capture_output(test_caps(Type, Unit, Name, Line,
+                                         Options, Body, Result, Key),
+                               Output, GlobalOptions),
+                Pairs0),
+        keysort(Pairs0, Pairs),
 	group_pairs_by_key(Pairs, Keyed),
 	(   Keyed == []
 	->  true
 	;   Keyed = [_-Results]
-	->  Results = [_Type+Result|_],
-	    report_result(Result, Options)          % consistent results
+	->  Results = [r(_Type,Result,Output)|_],
+	    report_result(Result, Output, Options)          % consistent results
 	;   pairs_values(Pairs, ResultByType),
-	    report_result(sto(Unit, Name, Line, ResultByType), Options)
+	    report_result(sto(Unit, Name, Line, ResultByType), "", Options)
 	)
     ).
 
@@ -897,155 +966,179 @@ test_caps(Type, Unit, Name, Line, Options, Body, Result, Key) :-
     run_test_6(Unit, Name, Line, Options, Body, Result),
     end_test(Unit, Name, Line, Type),
     result_to_key(Result, Key),
-    Key \== setup_failed.
+    Key \== setup_failed,
+    Key \== condition_failed.
 
+:- det(result_to_key/2).
 result_to_key(blocked(_, _, _, _), blocked).
-result_to_key(failure(_, _, _, How0), failure(How1)) :-
+result_to_key(failure(_, _, _, How0, _), failure(How1)) :-
     ( How0 = succeeded(_T) -> How1 = succeeded ; How0 = How1 ).
 result_to_key(success(_, _, _, Determinism, _), success(Determinism)).
 result_to_key(setup_failed(_,_,_), setup_failed).
+result_to_key(condition_failed(_,_,_), condition_failed).
+
+:- det(report_result/3).
+report_result(Result, Output, Options) :-
+    current_test_flag(test_options, GlobalOptions),
+    print_test_output(Result, Output, GlobalOptions),
+    report_result(Result, Options).
 
 report_result(blocked(Unit, Name, Line, Reason), _) :-
     !,
     assert(blocked(Unit, Name, Line, Reason)).
-report_result(failure(Unit, Name, Line, How), Options) :-
+report_result(failure(Unit, Name, Line, How, Time), Options) :-
     !,
-    failure(Unit, Name, Line, How, Options).
+    failure(Unit, Name, Line, How, Time, Options).
 report_result(success(Unit, Name, Line, Determinism, Time), Options) :-
     !,
     success(Unit, Name, Line, Determinism, Time, Options).
 report_result(setup_failed(_Unit, _Name, _Line), _Options).
+report_result(condition_failed(_Unit, _Name, _Line), _Options).
 report_result(sto(Unit, Name, Line, ResultByType), Options) :-
     assert(sto(Unit, Name, Line, ResultByType)),
     print_message(error, plunit(sto(Unit, Name, Line))),
     report_sto_results(ResultByType, Options).
 
 report_sto_results([], _).
-report_sto_results([Type+Result|T], Options) :-
+report_sto_results([r(Type,Result,Output)|T], Options) :-
+    print_test_output(Result, Output, Options),
     print_message(error, plunit(sto(Type, Result))),
     report_sto_results(T, Options).
 
+print_test_output(Result, Output, Options) :-
+    Output \== "",
+    option(output(on_failure), Options),
+    result_to_key(Result, Key),
+    Key \= success(_),
+    !,
+    ansi_format(code, '~N~s', [Output]).     % Use print_message(test(output), Output)?
+print_test_output(_, _, _).
+
 
 %!  run_test_6(+Unit, +Name, +Line, +Options, :Body, -Result) is det.
 %
+%   6th step  of the  tests.  Deals  with tests  that must  be ignored
+%   (blocked, conditions fails), setup and cleanup at the test level.
 %   Result is one of:
 %
-%           * blocked(Unit, Name, Line, Reason)
-%           * failure(Unit, Name, Line, How)
-%           * success(Unit, Name, Line, Determinism, Time)
-%           * setup_failed(Unit, Name, Line)
+%     - blocked(Unit, Name, Line, Reason)
+%     - condition_failed(Unit, Name, Line)
+%     - failure(Unit, Name, Line, How, Time)
+%       How is one of:
+%       - succeeded
+%       - Exception
+%       - cmp_error(Cmp, E)
+%       - wrong_answer(Cmp)
+%       - failed
+%       - no_exception
+%       - wrong_error(Expect, E)
+%       - wrong_answer(Expected, Bindings)
+%     - success(Unit, Name, Line, Determinism, Time)
+%     - setup_failed(Unit, Name, Line)
 
 run_test_6(Unit, Name, Line, Options, _Body,
 	   blocked(Unit, Name, Line, Reason)) :-
     option(blocked(Reason), Options),
     !.
+run_test_6(Unit, Name, Line, Options, _Body, Result) :-
+    unit_module(Unit, Module),
+    \+ condition(Module, test(Unit,Name,Line), Options),
+    !,
+    Result = condition_failed(Unit, Name, Line).
 run_test_6(Unit, Name, Line, Options, Body, Result) :-
+    option(setup(_Setup), Options),
+    !,
+    (   unit_module(Unit, Module),
+        setup(Module, test(Unit,Name,Line), Options)
+    ->  run_test_7(Unit, Name, Line, Options, Body, Result),
+        cleanup(Module, Options)
+    ;   Result = setup_failed(Unit, Name, Line)
+    ).
+run_test_6(Unit, Name, Line, Options, Body, Result) :-
+    unit_module(Unit, Module),
+    run_test_7(Unit, Name, Line, Options, Body, Result),
+    cleanup(Module, Options).
+
+%!  run_test_7(+Unit, +Name, +Line, +Options, :Body, -Result) is det.
+%
+%   This step  deals with the expected  outcome of the test.   It runs
+%   the  actual test  and then  compares  the result  to the  outcome.
+%   There are  two main categories:  dealing with a single  result and
+%   all results.
+
+run_test_7(Unit, Name, Line, Options, Body, Result) :-
+    option(true(Cmp), Options),			   % expected success
+    !,
+    unit_module(Unit, Module),
+    call_time(reify(call_det(Module:Body, Det), Result0), Time),
+    (   Result0 == true
+    ->  (   catch(Module:Cmp, E, true)
+        ->  (  var(E)
+            ->  Result = success(Unit, Name, Line, Det, Time)
+            ;   Result = failure(Unit, Name, Line, cmp_error(Cmp, E), Time)
+            )
+        ;   Result = failure(Unit, Name, Line, wrong_answer(Cmp), Time)
+        )
+    ;   Result0 == false
+    ->  Result = failure(Unit, Name, Line, failed, Time)
+    ;   Result0 = throw(E2)
+    ->  Result = failure(Unit, Name, Line, E2, Time)
+    ).
+run_test_7(Unit, Name, Line, Options, Body, Result) :-
+    option(fail, Options),                         % expected failure
+    !,
+    unit_module(Unit, Module),
+    call_time(reify(Module:Body, Result0), Time),
+    (   Result0 == true
+    ->  Result = failure(Unit, Name, Line, succeeded, Time)
+    ;   Result0 == false
+    ->  Result = success(Unit, Name, Line, true, Time)
+    ;   Result0 = throw(E)
+    ->  Result = failure(Unit, Name, Line, E, Time)
+    ).
+run_test_7(Unit, Name, Line, Options, Body, Result) :-
+    option(throws(Expect), Options),		   % Expected error
+    !,
+    unit_module(Unit, Module),
+    call_time(reify(Module:Body, Result0), Time),
+    (   Result0 == true
+    ->  Result = failure(Unit, Name, Line, no_exception, Time)
+    ;   Result0 == false
+    ->  Result = failure(Unit, Name, Line, failed, Time)
+    ;   Result0 = throw(E)
+    ->  (   match_error(Expect, E)
+        ->  Result = success(Unit, Name, Line, true, Time)
+        ;   Result = failure(Unit, Name, Line, wrong_error(Expect, E), Time)
+        )
+    ).
+run_test_7(Unit, Name, Line, Options, Body, Result) :-
     option(all(Answer), Options),                  % all(Bindings)
     !,
     nondet_test(all(Answer), Unit, Name, Line, Options, Body, Result).
-run_test_6(Unit, Name, Line, Options, Body, Result) :-
+run_test_7(Unit, Name, Line, Options, Body, Result) :-
     option(set(Answer), Options),                  % set(Bindings)
     !,
     nondet_test(set(Answer), Unit, Name, Line, Options, Body, Result).
-run_test_6(Unit, Name, Line, Options, Body, Result) :-
-    option(fail, Options),                         % fail
-    !,
-    unit_module(Unit, Module),
-    (   setup(Module, test(Unit,Name,Line), Options)
-    ->  statistics(runtime, [T0,_]),
-	(   catch(Module:Body, E, true)
-	->  (   var(E)
-	    ->  statistics(runtime, [T1,_]),
-		Time is (T1 - T0)/1000.0,
-		Result = failure(Unit, Name, Line, succeeded(Time)),
-		cleanup(Module, Options)
-	    ;   Result = failure(Unit, Name, Line, E),
-		cleanup(Module, Options)
-	    )
-	;   statistics(runtime, [T1,_]),
-	    Time is (T1 - T0)/1000.0,
-	    Result = success(Unit, Name, Line, true, Time),
-	    cleanup(Module, Options)
-	)
-    ;   Result = setup_failed(Unit, Name, Line)
-    ).
-run_test_6(Unit, Name, Line, Options, Body, Result) :-
-    option(true(Cmp), Options),
-    !,
-    unit_module(Unit, Module),
-    (   setup(Module, test(Unit,Name,Line), Options) % true(Binding)
-    ->  statistics(runtime, [T0,_]),
-	(   catch(call_det(Module:Body, Det), E, true)
-	->  (   var(E)
-	    ->  statistics(runtime, [T1,_]),
-		Time is (T1 - T0)/1000.0,
-		(   catch(Module:Cmp, E, true)
-		->  (   var(E)
-		    ->  Result = success(Unit, Name, Line, Det, Time)
-		    ;   Result = failure(Unit, Name, Line, cmp_error(Cmp, E))
-		    )
-		;   Result = failure(Unit, Name, Line, wrong_answer(Cmp))
-		),
-		cleanup(Module, Options)
-	    ;   Result = failure(Unit, Name, Line, E),
-		cleanup(Module, Options)
-	    )
-	;   Result = failure(Unit, Name, Line, failed),
-	    cleanup(Module, Options)
-	)
-    ;   Result = setup_failed(Unit, Name, Line)
-    ).
-run_test_6(Unit, Name, Line, Options, Body, Result) :-
-    option(throws(Expect), Options),
-    !,
-    unit_module(Unit, Module),
-    (   setup(Module, test(Unit,Name,Line), Options)
-    ->  statistics(runtime, [T0,_]),
-	(   catch(Module:Body, E, true)
-	->  (   var(E)
-	    ->  Result = failure(Unit, Name, Line, no_exception),
-		cleanup(Module, Options)
-	    ;   statistics(runtime, [T1,_]),
-		Time is (T1 - T0)/1000.0,
-		(   match_error(Expect, E)
-		->  Result = success(Unit, Name, Line, true, Time)
-		;   Result = failure(Unit, Name, Line, wrong_error(Expect, E))
-		),
-		cleanup(Module, Options)
-	    )
-	;   Result = failure(Unit, Name, Line, failed),
-	    cleanup(Module, Options)
-	)
-    ;   Result = setup_failed(Unit, Name, Line)
-    ).
 
 
 %!  non_det_test(+Expected, +Unit, +Name, +Line, +Options, +Body, -Result)
 %
 %   Run tests on non-deterministic predicates.
 
-nondet_test(Expected, Unit, Name, Line, Options, Body, Result) :-
+nondet_test(Expected, Unit, Name, Line, _Options, Body, Result) :-
     unit_module(Unit, Module),
     result_vars(Expected, Vars),
-    statistics(runtime, [T0,_]),
-    (   setup(Module, test(Unit,Name,Line), Options)
-    ->  (   catch(findall(Vars, Module:Body, Bindings), E, true)
-	->  (   var(E)
-	    ->  statistics(runtime, [T1,_]),
-		Time is (T1 - T0)/1000.0,
-		(   nondet_compare(Expected, Bindings, Unit, Name, Line)
-		->  Result = success(Unit, Name, Line, true, Time)
-		;   Result = failure(Unit, Name, Line, wrong_answer(Expected, Bindings))
-		),
-		cleanup(Module, Options)
-	    ;   Result = failure(Unit, Name, Line, E),
-		cleanup(Module, Options)
-	    )
-	)
-    ;   Result = setup_failed(Unit, Name, Line)
+    (   call_time(reify(findall(Vars, Module:Body, Bindings), Result0), Time)
+    ->  (   Result0 == true
+        ->  (   nondet_compare(Expected, Bindings, Unit, Name, Line)
+            ->  Result = success(Unit, Name, Line, true, Time)
+            ;   Result = failure(Unit, Name, Line, wrong_answer(Expected, Bindings), Time)
+            )
+        ;   Result0 = throw(E)
+        ->  Result = failure(Unit, Name, Line, E, Time)
+        )
     ).
 
-
 %!  result_vars(+Expected, -Vars) is det.
 %
 %   Create a term v(V1, ...) containing all variables at the left
@@ -1119,17 +1212,11 @@ match_error(Expect, Rec) :-
 %   reason. The condition handler is  similar,   but  failing is not
 %   considered an error.  Context is one of
 %
-%       * unit(Unit)
-%       If it is the setup handler for a unit
-%       * test(Unit,Name,Line)
-%       If it is the setup handler for a test
+%    - unit(Unit)
+%      If it is the setup handler for a unit
+%    - test(Unit,Name,Line)
+%      If it is the setup handler for a test
 
-setup(Module, Context, Options) :-
-    option(condition(Condition), Options),
-    option(setup(Setup), Options),
-    !,
-    setup(Module, Context, [condition(Condition)]),
-    setup(Module, Context, [setup(Setup)]).
 setup(Module, Context, Options) :-
     option(setup(Setup), Options),
     !,
@@ -1142,10 +1229,16 @@ setup(Module, Context, Options) :-
     ;   print_message(error, error(goal_failed(Setup), _)),
 	fail
     ).
-setup(Module, Context, Options) :-
-    option(condition(Setup), Options),
+setup(_,_,_).
+
+%!  condition(+Module, +Context, +Options) is semidet.
+%
+%   Evaluate the test or test unit condition.
+
+condition(Module, Context, Options) :-
+    option(condition(Cond), Options),
     !,
-    (   catch(call_ex(Module, Setup), E, true)
+    (   catch(call_ex(Module, Cond), E, true)
     ->  (   var(E)
 	->  true
 	;   print_message(error, plunit(error(condition, Context, E))),
@@ -1153,7 +1246,8 @@ setup(Module, Context, Options) :-
 	)
     ;   fail
     ).
-setup(_,_,_).
+condition(_, _, _).
+
 
 %!  call_ex(+Module, +Goal)
 %
@@ -1178,41 +1272,46 @@ cleanup(Module, Options) :-
     ;   print_message(warning, goal_failed(Cleanup, '(cleanup handler)'))
     ).
 
-success(Unit, Name, Line, Det, _Time, Options) :-
+success(Unit, Name, Line, Det, Time, Options) :-
     memberchk(fixme(Reason), Options),
     !,
     (   (   Det == true
 	;   memberchk(nondet, Options)
 	)
-    ->  progress(Unit, Name, nondet),
+    ->  progress(Unit, Name, fixme(passed), Time),
 	Ok = passed
-    ;   progress(Unit, Name, fixme),
+    ;   progress(Unit, Name, fixme(nondet), Time),
 	Ok = nondet
     ),
     flush_output(user_error),
     assert(fixme(Unit, Name, Line, Reason, Ok)).
-success(Unit, Name, Line, _, _, Options) :-
+success(Unit, Name, Line, _, Time, Options) :-
     failed_assertion(Unit, Name, Line, _,_,_,_),
     !,
-    failure(Unit, Name, Line, assertion, Options).
+    failure(Unit, Name, Line, assertion, Time, Options).
 success(Unit, Name, Line, Det, Time, Options) :-
     assert(passed(Unit, Name, Line, Det, Time)),
     (   (   Det == true
 	;   memberchk(nondet, Options)
 	)
-    ->  progress(Unit, Name, passed)
+    ->  progress(Unit, Name, passed, Time)
     ;   unit_file(Unit, File),
 	print_message(warning, plunit(nondet(File, Line, Name)))
     ).
 
-failure(Unit, Name, Line, _, Options) :-
+%!  failure(+Unit, +Name, +Line, +How, +Time, +Options) is det.
+%
+%   Test failed.  Report the error.
+
+failure(Unit, Name, Line, _, Time, Options) :-
     memberchk(fixme(Reason), Options),
     !,
-    progress(Unit, Name, failed),
+    progress(Unit, Name, fixme(failed), Time),
     assert(fixme(Unit, Name, Line, Reason, failed)).
-failure(Unit, Name, Line, E, Options) :-
-    report_failure(Unit, Name, Line, E, Options),
-    assert_cyclic(failed(Unit, Name, Line, E)).
+failure(Unit, Name, Line, E, Time, Options) :-
+    report_failure(Unit, Name, Line, E, Time, Options),
+    progress(Unit, Name, failed, Time),
+    assert_cyclic(failed(Unit, Name, Line, E, Time)).
 
 %!  assert_cyclic(+Term) is det.
 %
@@ -1284,7 +1383,7 @@ running_tests(Running) :-
 	    ), Running).
 
 
-%!  current_test(?Unit, ?Test, ?Line, ?Body, ?Options)
+%!  current_test(?Unit, ?Test, ?Line, ?Body, ?Options) is nondet.
 %
 %   True when a test with the specified properties is loaded.
 
@@ -1292,6 +1391,14 @@ current_test(Unit, Test, Line, Body, Options) :-
     current_unit(Unit, Module, _Supers, _UnitOptions),
     Module:'unit test'(Test, Line, Options, Body).
 
+%!  current_test_unit(?Unit, ?Options) is nondet.
+%
+%   True when a Unit is a current unit test declared with Options.
+
+current_test_unit(Unit, UnitOptions) :-
+    current_unit(Unit, _Module, _Supers, UnitOptions).
+
+
 :- meta_predicate count(0, -).
 count(Goal, Count) :-
     aggregate_all(count, Goal, Count).
@@ -1301,11 +1408,11 @@ count(Goal, Count) :-
 %   True if there are no failures, otherwise false.
 
 test_summary(Unit, Summary) :-
-    count(failed(Unit, _0Test, _0Line, _Reason), Failed),
+    count(failed(Unit, _0Test, _0Line, _Reason, _0Time), Failed),
     count(failed_assertion(Unit, _0Test, _0Line,
 			   _ALoc, _STO, _0Reason, _Goal), FailedAssertion),
     count(sto(Unit, _0Test, _0Line, _Results), STO),
-    count(passed(Unit, _0Test, _0Line, _Det, _Time), Passed),
+    count(passed(Unit, _0Test, _0Line, _Det, _0Time), Passed),
     count(blocked(Unit, _0Test, _0Line, _0Reason), Blocked),
     Summary = plunit{passed:Passed,
 		     failed:Failed,
@@ -1368,7 +1475,7 @@ report_blocked :-
 report_blocked.
 
 report_failed :-
-    number_of_clauses(failed/4, N),
+    number_of_clauses(failed/5, N),
     info(plunit(failed(N))).
 
 report_failed_assertions :-
@@ -1395,16 +1502,17 @@ fixme(How, Tuples, Count) :-
     length(Tuples, Count).
 
 
-report_failure(Unit, Name, _, assertion, _) :-
+report_failure(Unit, Name, _, assertion, Time, _) :-
     !,
-    progress(Unit, Name, assertion).
-report_failure(Unit, Name, Line, Error, _Options) :-
+    progress(Unit, Name, assertion, Time).
+report_failure(Unit, Name, Line, Error, _Time, _Options) :-
     print_message(error, plunit(failed(Unit, Name, Line, Error))).
 
 
-%!  test_report(What) is det.
+%!  test_report(+What) is det.
 %
-%   Produce reports on test results after the run.
+%   Produce reports on test  results  after   the  run.  Currently  only
+%   supports `fixme` for What.
 
 test_report(fixme) :-
     !,
@@ -1479,8 +1587,10 @@ info(Term) :-
     message_level(Level),
     print_message(Level, Term).
 
-progress(Unit, Name, Result) :-
-    print_message(information, plunit(progress(Unit, Name, Result))).
+progress(Unit, Name, Result, Time) :-
+    flag(plunit_test, N, N+1),
+    test_count(Total),
+    print_message(information, plunit(progress(Unit, Name, Result, N/Total, Time))).
 
 message_level(Level) :-
     current_test_flag(test_options, Options),
@@ -1519,7 +1629,7 @@ message(error(plunit(incompatible_options, Tests), _)) -->
 
 					% Unit start/end
 :- if(swi).
-message(plunit(progress(_Unit, _Name, Result))) -->
+message(plunit(progress(_Unit, _Name, Result, _N_T, _Time))) -->
     [ at_same_line ], result(Result), [flush].
 message(plunit(begin(Unit))) -->
     [ 'PL-Unit: ~w '-[Unit], flush ].
@@ -1529,7 +1639,7 @@ message(plunit(end(_Unit, Summary))) -->
     ->  [ ' passed' ]
     ;   [ ansi(error, '**FAILED', []) ]
     ),
-    [ ' ~3f sec'-[Summary.time] ].
+    [ ' ~3f sec'-[Summary.time.cpu] ].
 :- else.
 message(plunit(begin(Unit))) -->
     [ 'PL-Unit: ~w '-[Unit]/*, flush-[]*/ ].
@@ -1660,11 +1770,12 @@ unqualify(M:Goal, _, Goal) :-
     !.
 unqualify(Goal, _, Goal).
 
-result(passed)    --> ['.'-[]].
-result(nondet)    --> ['+'-[]].
-result(fixme)     --> ['!'-[]].
-result(failed)    --> ['-'-[]].
-result(assertion) --> ['A'-[]].
+result(passed)        --> ['.'-[]].
+result(nondet)        --> ['+'-[]].
+result(fixme(passed)) --> ['*'-[]].
+result(fixme(failed)) --> ['!'-[]].
+result(failed)        --> ['-'-[]].
+result(assertion)     --> ['A'-[]].
 
 :- endif.
 					% Setup/condition errors
@@ -1676,9 +1787,9 @@ message(plunit(error(Where, Context, Exception))) -->
 					% STO messages
 message(plunit(sto(Unit, Name, Line))) -->
     { unit_file(Unit, File) },
-       locationprefix(File:Line),
-       test_name(Name),
-       [' is subject to occurs check (STO): '-[] ].
+    locationprefix(File:Line),
+    test_name(Name),
+    [' is subject to occurs check (STO): '-[] ].
 message(plunit(sto(Type, Result))) -->
     sto_type(Type),
     sto_result(Result).
@@ -1715,8 +1826,8 @@ sto_type(finite_trees) -->
 
 sto_result(success(_Unit, _Name, _Line, Det, Time)) -->
     det(Det),
-    [ ' success in ~2f seconds'-[Time] ].
-sto_result(failure(_Unit, _Name, _Line, How)) -->
+    [ ' success in ~3f seconds'-[Time.cpu] ].
+sto_result(failure(_Unit, _Name, _Line, How, _Time)) -->
     failure(How).
 
 det(true) -->
@@ -1813,7 +1924,7 @@ failure(Error) -->
     [ 'received error: ~w'-[Message] ].
 :- endif.
 failure(Why) -->
-    [ '~p~n'-[Why] ].
+    [ '~p'-[Why] ].
 
 fixme_message([]) --> [].
 fixme_message([fixme(Unit, _Name, Line, Reason, How)|T]) -->
