@@ -1,7 +1,7 @@
 /*
     Author:        Eshel Yaron
     E-mail:        eshel@swi-prolog.org
-    Copyright (c)  2022, SWI-Prolog Solutions b.v.
+    Copyright (c)  2022-2023, SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -70,7 +70,12 @@
             sweep_interrupt/0,
             sweep_string_to_atom/2,
             sweep_file_path_in_library/2,
-            sweep_file_missing_dependencies/2
+            sweep_file_missing_dependencies/2,
+            sweep_format_head/2,
+            sweep_format_term/2,
+            sweep_current_functors/2,
+            sweep_term_search/2,
+            sweep_terms_at_point/2
           ]).
 
 :- use_module(library(pldoc)).
@@ -353,23 +358,15 @@ sweep_predicate_apropos(Query0, Matches) :-
             Tail).
 
 sweep_predicate_location_(H, Path, Line) :-
-    predicate_property(H, file(Path0)),
-    predicate_property(H, line_count(Line)),
+    xref_defined(Path0, H, How),
+    xref_definition_line(How, Line),
     !,
     atom_string(Path0, Path).
 sweep_predicate_location_(H, Path, Line) :-
-    (   xref_defined(Path0, H, How),
-        xref_definition_line(How, Line)
-    ->  true
-    ;   xref_defined(Path0, H, _), Line = []
-    ),
+    predicate_property(H, file(Path0)),
+    predicate_property(H, line_count(Line)),
     atom_string(Path0, Path).
 
-sweep_predicate_location_(M, H, Path, Line) :-
-    predicate_property(M:H, file(Path0)),
-    predicate_property(M:H, line_count(Line)),
-    !,
-    atom_string(Path0, Path).
 sweep_predicate_location_(M, H, Path, Line) :-
     (   xref_defined(Path0, M:H, How),
         xref_definition_line(How, Line)
@@ -377,9 +374,12 @@ sweep_predicate_location_(M, H, Path, Line) :-
     ;   xref_defined(Path0, H, How),
         xref_definition_line(How, Line),
         xref_module(Path0, M)
-    ->  true
-    ;   xref_defined(Path0, M:H, _), Line = []
     ),
+    !,
+    atom_string(Path0, Path).
+sweep_predicate_location_(M, H, Path, Line) :-
+    predicate_property(M:H, file(Path0)),
+    predicate_property(M:H, line_count(Line)),
     atom_string(Path0, Path).
 
 
@@ -757,13 +757,19 @@ sweep_beginning_of_last_predicate(Start, Next) :-
                 L < Start
             ),
             Ls),
-    reverse(Ls, [Next|_]).
+    sort(Ls, Sorted),
+    reverse(Sorted, [Next|_]).
 
 sweep_beginning_of_next_predicate(Start, Next) :-
     sweep_source_id(Path),
     xref_source(Path, [comments(store)]),
-    xref_defined(Path, _, H), xref_definition_line(H, Next),
-    Start < Next.
+    findall(L,
+            (   xref_defined(Path, _, H),
+                xref_definition_line(H, L),
+                Start < L
+            ),
+            Ls),
+    sort(Ls, [Next|_]).
 
 sweep_source_id(Path) :-
     sweep_main_thread,
@@ -942,3 +948,143 @@ dep_import(Path, Kind, PI0) -->
     {   term_string(PI0, PI)
     },
     [[Path, PI, Kind]].
+
+
+sweep_format_head([F0|A], R) :-
+    atom_string(F, F0),
+    pi_head(F/A, H),
+    sweep_current_module(M),
+    sweep_format_predicate(M, 0, H, R).
+
+sweep_format_term([F0,N,P], [S|SP]) :-
+    atom_string(F, F0),
+    pi_head(F/N, H),
+    length(NamedArgs, N),
+    maplist(=('$VAR'('_')), NamedArgs),
+    H =.. [F|NamedArgs],
+    term_string(H, S, [quoted(true),
+                       character_escapes(true),
+                       spacing(next_argument),
+                       numbervars(true),
+                       priority(P)]),
+    term_string(_, S, [subterm_positions(SP)]).
+
+sweep_current_functors(A0, Col) :-
+    (   A0 == []
+    ->  true
+    ;   A = A0
+    ),
+    findall([F|A],
+            (   current_functor(F0, A),
+                atom(F0),
+                atom_string(F0, F)
+            ),
+            Col).
+
+sweep_term_search([Path0,TermString,GoalString], Res) :-
+    term_string(Term, TermString, [variable_names(TermVarNames)]),
+    term_string(Goal, GoalString, [variable_names(GoalVarNames)]),
+    maplist({GoalVarNames}/[TermVarName]>>ignore(memberchk(TermVarName, GoalVarNames)),
+            TermVarNames),
+    atom_string(Path, Path0),
+    setup_call_cleanup(prolog_open_source(Path, Stream),
+                       sweep_search_stream(Stream, Term, Goal, Res),
+                       prolog_close_source(Stream)).
+
+sweep_search_stream(Stream, Term, Goal, Res) :-
+    prolog_read_source_term(Stream, Term0, _, [subterm_positions(TermPos)]),
+    sweep_search_stream_(Term0, TermPos, Stream, Term, Goal, Res).
+
+sweep_search_stream_(end_of_file, _, _, _, _, []) :-
+    !.
+sweep_search_stream_(Term0, TermPos, Stream, Term, Goal, Res) :-
+    findall([HS|HE],
+            sweep_match_term(TermPos, Term0, Term, Goal, HS, HE),
+            Res,
+            Tail),
+    sweep_search_stream(Stream, Term, Goal, Tail).
+
+sweep_match_term(Pos, Term0, Term, Goal, From, To) :-
+    compound(Pos),
+    Pos \= parentheses_term_position(_, _, _),
+    arg(1, Pos, From),
+    arg(2, Pos, To),
+    subsumes_term(Term, Term0),
+    \+ \+ (   Term = Term0,
+              Goal
+          ).
+sweep_match_term(brace_term_position(_, _, Arg), {Term0}, Term, Goal, From, To) :-
+    sweep_match_term(Arg, Term0, Term, Goal, From, To).
+sweep_match_term(list_position(_, _, Elms, _), Term0, Term, Goal, From, To) :-
+    nth0(I, Elms, Elm),
+    nth0(I, Term0, Term1),
+    sweep_match_term(Elm, Term1, Term, Goal, From, To).
+sweep_match_term(list_position(_, _, _, Tail), Term0, Term, Goal, From, To) :-
+    list_tail(Term0, Term1),
+    sweep_match_term(Tail, Term1, Term, Goal, From, To).
+sweep_match_term(term_position(_, _, _, _, SubPos), Term0, Term, Goal, From, To) :-
+    nth1(I, SubPos, Sub),
+    arg(I, Term0, Term1),
+    sweep_match_term(Sub, Term1, Term, Goal, From, To).
+sweep_match_term(dict_position(_, _, _, _, KeyValuePosList), Term0, Term, Goal, From, To) :-
+    member(key_value_position(_, _, _, _, Key, _, ValuePos), KeyValuePosList),
+    get_dict(Key, Term0, Term1),
+    sweep_match_term(ValuePos, Term1, Term, Goal, From, To).
+sweep_match_term(parentheses_term_position(_, _, ContentPos), Term0, Term, Goal, From, To) :-
+    sweep_match_term(ContentPos, Term0, Term, Goal, From, To).
+sweep_match_term(quasi_quotation_position(_, _, SyntaxTerm, SyntaxPos, _), _, Term, Goal, From, To) :-
+    sweep_match_term(SyntaxPos, SyntaxTerm, Term, Goal, From, To).
+
+list_tail([_|T0], T) :- nonvar(T0), T0 = [_|_], !, list_tail(T0, T).
+list_tail([_|T], T).
+
+sweep_terms_at_point([String, Start, Point], Res) :-
+    (   sweep_source_id(Path0),
+        atom_string(Path, Path0),
+        findall(Op, xref_op(Path, Op), Ops),
+        (   xref_module(Path, Module)
+        ->  true
+        ;   Module = user
+        )
+    ->  true
+    ;   Module = user, Ops = []
+    ),
+    with_buffer_stream(
+        Stream,
+        String,
+        (   ignore((   nonvar(Path),
+                       set_stream(Stream, file_name(Path))
+                   )),
+            read_source_term_at_location(Stream, _,
+                                         [module(Module),
+                                          operators(Ops),
+                                          subterm_positions(SubPos)]),
+            findall([Beg|End],
+                    sweep_terms_at_point_(SubPos, Start, Point, Beg, End),
+                    Res)
+        )).
+
+sweep_terms_at_point_(SubPos, Start, Point, Beg, End) :-
+    SubPos \= parentheses_term_position(_, _, _),
+    arg(1, SubPos, Beg0),
+    arg(2, SubPos, End0),
+    Beg0 =< Point,
+    Point =< End0,
+    Beg is Beg0 + Start,
+    End is End0 + Start.
+sweep_terms_at_point_(list_position(_, _, Elms, _), Start, Point, Beg, End) :-
+    member(SubPos, Elms),
+    sweep_terms_at_point_(SubPos, Start, Point, Beg, End).
+sweep_terms_at_point_(list_position(_, _, _, SubPos), Start, Point, Beg, End) :-
+    SubPos \== none,
+    sweep_terms_at_point_(SubPos, Start, Point, Beg, End).
+sweep_terms_at_point_(term_position(_, _, _, _, Args), Start, Point, Beg, End) :-
+    member(SubPos, Args),
+    sweep_terms_at_point_(SubPos, Start, Point, Beg, End).
+sweep_terms_at_point_(dict_position(_, _, _, _, KeyValuePosList), Start, Point, Beg, End) :-
+    member(key_value_position(_, _, _, _, _, _, SubPos), KeyValuePosList),
+    sweep_terms_at_point_(SubPos, Start, Point, Beg, End).
+sweep_terms_at_point_(parentheses_term_position(_, _, SubPos), Start, Point, Beg, End) :-
+    sweep_terms_at_point_(SubPos, Start, Point, Beg, End).
+sweep_terms_at_point_(quasi_quotation_position(_, _, _, SubPos, _), Start, Point, Beg, End) :-
+    sweep_terms_at_point_(SubPos, Start, Point, Beg, End).
