@@ -5,6 +5,10 @@
 (remove-hook 'flymake-diagnostic-functions
              #'flymake-proc-legacy-flymake)
 
+(add-hook 'sweeprolog-mode-hook (lambda ()
+                                  (setq-local indent-tabs-mode nil
+                                              inhibit-message t)))
+
 (defconst sweeprolog-tests-greeting
   "Hello from Elisp from Prolog from Elisp from Prolog from Elisp!")
 
@@ -17,7 +21,7 @@
     (cdr sol)))
 
 (defun sweeprolog-tests-greet-1 ()
-  (message sweeprolog-tests-greeting))
+  sweeprolog-tests-greeting)
 
 (ert-deftest elisp->prolog->elisp->prolog->elisp ()
   "Tests calling Elisp from Prolog from Elisp from Prolog from Elisp."
@@ -74,6 +78,59 @@ foo(Baz) :- baz.
     (should (not (sweeprolog-beginning-of-next-top-term)))
     (should (= (point) 19))))
 
+(ert-deftest terms-at-point ()
+  "Test `sweeprolog-term-search'."
+  (let ((temp (make-temp-file "sweeprolog-terms-at-point-test"
+                              nil
+                              "pl"
+                              "
+recursive(Var) :-
+    (   true
+    ->  recursive(Bar)
+    ;   var(Baz)
+    *-> Bar is foo
+    ).
+")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (should (equal (sweeprolog-terms-at-point 81)
+                '("Bar"
+                  "Bar is foo"
+                  "var(Baz)
+    *-> Bar is foo" "true
+    ->  recursive(Bar)
+    ;   var(Baz)
+    *-> Bar is foo"
+    "recursive(Var) :-
+    (   true
+    ->  recursive(Bar)
+    ;   var(Baz)
+    *-> Bar is foo
+    )")))))
+
+(ert-deftest term-search ()
+  "Test `sweeprolog-term-search'."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+bar(bar(bar), bar{bar:bar}, [bar,bar|bar]).
+foo([Bar|Baz]).
+")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (point-min))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 10))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 24))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 31))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 35))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 39))))
+
 (ert-deftest beginning-of-next-top-term-header ()
   "Test finding the beginning of the first top term."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -177,6 +234,26 @@ foo(Foo) :- bar.
                    '(sweeprolog-undefined-default-face
                      sweeprolog-body-default-face)))))
 
+(ert-deftest insert-term-with-holes ()
+  "Test `sweeprolog-insert-term-with-holes'."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (sweeprolog-insert-term-with-holes ":-" 2)
+    (call-interactively #'kill-region)
+    (sweeprolog-insert-term-with-holes "foo" 3)
+    (call-interactively #'kill-region)
+    (sweeprolog-insert-term-with-holes "bar" 0)
+    (call-interactively #'kill-region)
+    (sweeprolog-insert-term-with-holes ";" 2)
+    (call-interactively #'kill-region)
+    (sweeprolog-insert-term-with-holes "->" 2)
+    (should (string= (buffer-string)
+                     "foo(bar, (_->_;_), _):-_."))))
+
 (ert-deftest plunit-testset-skeleton ()
   "Tests inserting PlUnit test-set blocks."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -587,7 +664,7 @@ foo.
 foo :- Body.
 "))))
 
-(ert-deftest dwim-next-clause-module-qualified-cdg ()
+(ert-deftest dwim-next-clause-module-qualified-dcg ()
   "Tests inserting new module-qualified DCG non-terminal."
   (let ((temp (make-temp-file "sweeprolog-test"
                               nil
@@ -608,6 +685,30 @@ spam:foo --> Body.
 "
                      ))))
 
+(ert-deftest dwim-next-clause-args ()
+  "Tests inserting new clause with arguments."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+%!  foo(+Bar) is det.
+
+foo(bar) :- bar.
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (point-max))
+    (sweeprolog-insert-term-dwim)
+    (should (string= (buffer-string)
+                              "
+%!  foo(+Bar) is det.
+
+foo(bar) :- bar.
+foo(Bar) :- Body.
+
+"))))
+
 (ert-deftest dwim-next-clause-module-qualified ()
   "Tests inserting new module-qualified clause."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -762,6 +863,38 @@ bar(X) :- permutation(X, [1,2,3]).
 "
                      ))))
 
+(ert-deftest update-dependencies-autoload-from-package ()
+  "Tests making implicit autoloads from a package explicit."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+bar(X) :- http_open(X, X, X).
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (call-interactively #'sweeprolog-update-dependencies)
+    (should (string= (buffer-string)
+                     "
+:- module(foo, [bar/1]).
+
+/** <module> Foo
+
+*/
+
+:- autoload(library(http/http_open), [ http_open/3
+                                     ]).
+
+bar(X) :- http_open(X, X, X).
+"))))
+
 (ert-deftest update-dependencies ()
   "Tests making implicit autoloads explicit."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -988,6 +1121,53 @@ scasp_and_show(Q, Model, Tree) :-
     (sweeprolog-end-of-top-term)
     (should (= (point) 252))))
 
+(ert-deftest beginning-of-predicate-definition-near-bob ()
+  "Test finding the beginning of the first predicate definition."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "foo :- bar.")
+    (goto-char (point-min))
+    (sweeprolog-beginning-of-predicate-at-point)
+    (should (= (point) (point-min)))))
+
+(ert-deftest align-spacs-in-line-comment ()
+  "Test using `sweeprolog-align-spaces' in a line comment."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "
+%!  foo is det.
+%
+%")
+    (sweeprolog-align-spaces)
+    (should (string= (buffer-string)
+                     "
+%!  foo is det.
+%
+%   "))))
+
+(ert-deftest auto-fill-pldoc-comments ()
+  "Test writing PlDoc comments with `auto-fill-mode' enable."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (auto-fill-mode)
+    (seq-do (lambda (c)
+              (let ((last-command-event c))
+                (call-interactively #'self-insert-command)))
+            "
+%!  foobar is det.
+%
+%   Nam vestibulum accumsan nisl.  Donec pretium posuere tellus.  Aenean in sem ac leo mollis blandit.  Nam a sapien.  Proin quam nisl, tincidunt et, mattis eget, convallis nec, purus.
+"
+            )
+    (should (string= (buffer-string)
+                     "
+%!  foobar is det.
+%
+%   Nam vestibulum accumsan nisl.  Donec pretium posuere tellus.
+%   Aenean in sem ac leo mollis blandit.  Nam a sapien.  Proin quam
+%   nisl, tincidunt et, mattis eget, convallis nec, purus.
+"))))
+
 (ert-deftest electric-layout ()
   "Test `sweeprolog-electric-layout-mode'."
   (with-temp-buffer
@@ -1051,6 +1231,25 @@ test_bindings(Name-Value) -->
 "
                      ))))
 
+(ert-deftest infer-indent-style ()
+  "Test inferring indentation style from buffer contents."
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "
+foo :-
+  bar.")
+    (sweeprolog-infer-indent-style)
+    (should (= sweeprolog-indent-offset 2))
+    (should (not indent-tabs-mode)))
+  (with-temp-buffer
+    (sweeprolog-mode)
+    (insert "
+foo :-
+\tbar.")
+    (sweeprolog-infer-indent-style)
+    (should (= sweeprolog-indent-offset tab-width))
+    (should indent-tabs-mode)))
+
 (ert-deftest custom-indentation ()
   "Test forcefully setting custom indentation levels."
   (with-temp-buffer
@@ -1072,7 +1271,8 @@ foo :-
   (with-temp-buffer
     (sweeprolog-mode)
     (insert given)
-    (indent-region-line-by-line (point-min) (point-max))
+    (let ((inhibit-message t))
+      (indent-region-line-by-line (point-min) (point-max)))
     (should (string= (buffer-substring-no-properties (point-min) (point-max))
                      expected))))
 
