@@ -161,6 +161,7 @@ class Prolog
     this.objects = {};			// id --> Object
     this.object_ids = new WeakMap();	// objec --> id
     this.next_object_id = 0;
+    this.open_queries = [];		// Stack with open queries
 
     this.__set_foreign_constants();
     this.__bind_foreign_functions();
@@ -173,17 +174,18 @@ class Prolog
   { let argv0 = this.args || [];
     argv0.unshift("swipl");
     let argv = argv0.map(function(arg) {
-        return this.module.allocate(
-            this.module.intArrayFromString(arg),
-            'i8', this.module.ALLOC_NORMAL);
+	return this.module.allocate(
+	    this.module.intArrayFromString(arg),
+	    'i8', this.module.ALLOC_NORMAL);
     }, this);
     var ptr = this.module._malloc(argv.length * 4);
     argv.forEach(function(arg, i) {
-        this.module.setValue(ptr + i * 4, arg, '*');
+	this.module.setValue(ptr + i * 4, arg, '*');
     }, this);
     if (!this.bindings.PL_initialise(argv.length, ptr)) {
-        throw new Error('SWI-Prolog initialisation failed.');
+	throw new Error('SWI-Prolog initialisation failed.');
     }
+    this.MODULE_user = this.new_module("user");
     this.call("set_prolog_flag(color_term, false).");
     this.call("set_prolog_flag(debug_on_error, false)");
     this.call("use_module(library(wasm))");
@@ -295,83 +297,83 @@ class Prolog
   __bind_foreign_functions()
   { this.bindings =
     { _PL_streams: this.module.cwrap(
-        '_PL_streams', 'number', []),
+	'_PL_streams', 'number', []),
       PL_functor_arity: this.module.cwrap(
-        'PL_functor_arity', 'number', ['number']),
+	'PL_functor_arity', 'number', ['number']),
       PL_functor_name: this.module.cwrap(
-        'PL_functor_name', 'number', ['number']),
+	'PL_functor_name', 'number', ['number']),
       PL_get_functor: this.module.cwrap(
-        'PL_get_functor', 'number', ['number', 'number']),
+	'PL_get_functor', 'number', ['number', 'number']),
       PL_get_chars: this.module.cwrap(
-        'PL_get_chars', 'number', ['number', 'number', 'number']),
+	'PL_get_chars', 'number', ['number', 'number', 'number']),
       PL_get_arg: this.module.cwrap(
-        'PL_get_arg', 'number', ['number', 'number', 'number']),
+	'PL_get_arg', 'number', ['number', 'number', 'number']),
       PL_get_int64: this.module.cwrap(
-        'PL_get_int64', 'number', ['number', 'number']),
+	'PL_get_int64', 'number', ['number', 'number']),
       PL_get_float: this.module.cwrap(
-        'PL_get_float', 'number', ['number', 'number']),
+	'PL_get_float', 'number', ['number', 'number']),
       PL_put_chars: this.module.cwrap(
-        'PL_put_chars', 'number', ['number', 'number', 'number', 'number']),
+	'PL_put_chars', 'number', ['number', 'number', 'number', 'number']),
       put_bytes: this.module.cwrap(
-        'PL_put_chars', 'number', ['number', 'number', 'number', 'array']),
+	'PL_put_chars', 'number', ['number', 'number', 'number', 'array']),
       PL_put_atom: this.module.cwrap(
-        'PL_put_atom', 'number', ['number']),
+	'PL_put_atom', 'number', ['number']),
       PL_put_variable: this.module.cwrap(
-        'PL_put_variable', 'number', ['number']),
+	'PL_put_variable', 'number', ['number']),
       PL_unify: this.module.cwrap(
-        'PL_unify', 'number', ['number', 'number']),
+	'PL_unify', 'number', ['number', 'number']),
       PL_is_string: this.module.cwrap(
-        'PL_is_string', 'number', ['number']),
+	'PL_is_string', 'number', ['number']),
       PL_is_variable: this.module.cwrap(
-        'PL_is_variable', 'number', ['number']),
+	'PL_is_variable', 'number', ['number']),
       PL_term_type: this.module.cwrap(
-        'PL_term_type', 'number', ['number']),
+	'PL_term_type', 'number', ['number']),
       PL_get_list: this.module.cwrap(
-        'PL_get_list', 'number', ['number', 'number', 'number']),
+	'PL_get_list', 'number', ['number', 'number', 'number']),
       PL_get_nil: this.module.cwrap(
-        'PL_get_nil', 'number', ['number']),
+	'PL_get_nil', 'number', ['number']),
       PL_initialise: this.module.cwrap(
-        'PL_initialise', 'number', ['number', 'number']),
+	'PL_initialise', 'number', ['number', 'number']),
       PL_new_atom: this.module.cwrap(
-        'PL_new_atom', 'number', ['string']),
+	'PL_new_atom', 'number', ['string']),
       PL_register_atom: this.module.cwrap(
-        'PL_register_atom', null, ['number']),
+	'PL_register_atom', null, ['number']),
       PL_unregister_atom: this.module.cwrap(
-        'PL_unregister_atom', null, ['number']),
+	'PL_unregister_atom', null, ['number']),
       PL_new_module: this.module.cwrap(
-        'PL_new_module', 'number', ['number']),
+	'PL_new_module', 'number', ['number']),
       PL_new_functor: this.module.cwrap(
-        'PL_new_functor', 'number', ['number', 'number']),
+	'PL_new_functor', 'number', ['number', 'number']),
       PL_new_term_ref: this.module.cwrap(
-        'PL_new_term_ref', 'number', []),
+	'PL_new_term_ref', 'number', []),
       PL_new_term_refs: this.module.cwrap(
-        'PL_new_term_refs', 'number', ['number']),
+	'PL_new_term_refs', 'number', ['number']),
       PL_copy_term_ref: this.module.cwrap(
-        'PL_copy_term_ref', 'number', ['number']),
+	'PL_copy_term_ref', 'number', ['number']),
       PL_reset_term_refs: this.module.cwrap(
-        'PL_reset_term_refs', null, ['number']),
+	'PL_reset_term_refs', null, ['number']),
       PL_put_functor: this.module.cwrap(
-        'PL_put_functor', 'number', ['number', 'number']),
+	'PL_put_functor', 'number', ['number', 'number']),
       PL_put_integer: this.module.cwrap(
-        'PL_put_integer', 'number', ['number', 'number']),
+	'PL_put_integer', 'number', ['number', 'number']),
       PL_put_float: this.module.cwrap(
-        'PL_put_float', 'number', ['number', 'number']),
+	'PL_put_float', 'number', ['number', 'number']),
       PL_put_nil: this.module.cwrap(
-        'PL_put_nil', 'number', []),
+	'PL_put_nil', 'number', []),
       PL_cons_functor_v: this.module.cwrap(
-        'PL_cons_functor_v', 'number', ['number', 'number', 'number']),
+	'PL_cons_functor_v', 'number', ['number', 'number', 'number']),
       PL_cons_list: this.module.cwrap(
-        'PL_cons_list', 'number', ['number', 'number', 'number']),
+	'PL_cons_list', 'number', ['number', 'number', 'number']),
       PL_put_dict: this.module.cwrap(
-        'PL_put_dict', 'number', ['number','number','number','number','number']),
+	'PL_put_dict', 'number', ['number','number','number','number','number']),
       PL_put_term_from_chars: this.module.cwrap(
-        'PL_put_term_from_chars', 'number',['number','number','number','string']),
+	'PL_put_term_from_chars', 'number',['number','number','number','string']),
       PL_put_term: this.module.cwrap(
-        'PL_put_term', 'number', ['number', 'number']),
+	'PL_put_term', 'number', ['number', 'number']),
       PL_write_term: this.module.cwrap(
-        'PL_write_term', 'number', ['number', 'number', 'number', 'number']),
+	'PL_write_term', 'number', ['number', 'number', 'number', 'number']),
       PL_call: this.module.cwrap(
-        'PL_call', 'number', ['number', 'number']),
+	'PL_call', 'number', ['number', 'number']),
       PL_open_foreign_frame: this.module.cwrap(
 	'PL_open_foreign_frame', 'number', []),
       PL_close_foreign_frame: this.module.cwrap(
@@ -390,18 +392,20 @@ class Prolog
 	'PL_cut_query', 'number', ['number']),
       PL_exception: this.module.cwrap(
 	'PL_exception', 'number', ['number']),
+      PL_raise_exception: this.module.cwrap(
+	'PL_raise_exception', 'number', ['number']),
       WASM_ttymode: this.module.cwrap(
-        'WASM_ttymode', 'number', []),
+	'WASM_ttymode', 'number', []),
       WASM_yield_request: this.module.cwrap(
-        'WASM_yield_request', 'number', []),
+	'WASM_yield_request', 'number', []),
       WASM_set_yield_result: this.module.cwrap(
-        'WASM_set_yield_result', 'number', ['number']),
+	'WASM_set_yield_result', 'number', ['number']),
       WASM_variable_id: this.module.cwrap(
-        'WASM_variable_id', 'number', ['number']),
+	'WASM_variable_id', 'number', ['number']),
       js_unify_obj: this.module.cwrap(
-        'js_unify_obj', 'number', ['number', 'number']),
+	'js_unify_obj', 'number', ['number', 'number']),
       js_get_obj: this.module.cwrap(
-        'js_get_obj', 'number', ['number'])
+	'js_get_obj', 'number', ['number'])
     };
   }
 
@@ -422,12 +426,13 @@ class Prolog
       { return this.__call_yieldable(goal, opts);
       } else
       { return this.with_frame(function()
-        { const term = this.new_term_ref();
+	{ const term = this.new_term_ref();
 
 	  if ( !this.chars_to_term(goal, term) )
 	    throw new Error('Query has a syntax error: ' + query);
 
-	  const module = opts.module ? this.new_module(opts.module) : 0;
+	  const module = opts.module ? this.new_module(opts.module)
+				     : this.MODULE_user;
 	  return !!this.bindings.PL_call(term, module);
 	});
       }
@@ -528,7 +533,7 @@ class Prolog
  */
 
   consult(...args)
-  { return this.forEach("load_files(Files)", {Files:args});
+  { return this.forEach("load_files(user:Files)", {Files:args});
   }
 
   load_string(s, id)
@@ -538,7 +543,7 @@ class Prolog
     }
     return this.forEach("setup_call_cleanup("+
 			   "open_string(S, _In),"+
-			   "load_files(Id, [stream(_In)]),"+
+			   "load_files(user:Id, [stream(_In)]),"+
 			   "close(_In))",
 			{S:new this.String(s), Id:id});
   }
@@ -546,12 +551,17 @@ class Prolog
   async load_scripts()
   { const prolog = this;
     const scripts = document.querySelectorAll("script[type='text/prolog']");
+    const loaded = [];
 
     for(let i = 0; i<scripts.length; i++)
     { const s = scripts[i];
+      const name = `/script/${s.id||s.name||i+1}`;
 
-      await prolog.load_string(s.text, `/script/${s.id||s.name||i+1}`);
+      await prolog.load_string(s.text, name);
+      loaded.push(name);
     }
+
+    return name;
   }
 
 
@@ -570,7 +580,7 @@ class Prolog
   url_properties(url)
   { return fetch(url, {method: 'HEAD'}).then((r) =>
     { if ( r.status == 200 )
-      { const size = parseInt(r.headers.get("content-length"));
+      { let size = parseInt(r.headers.get("content-length"));
 	const mods = r.headers.get("last-modified");
 	const time = Date.parse(mods) || 0;
 
@@ -693,7 +703,7 @@ class Prolog
     this.put_chars(av+0, goal);
     this.toProlog(input, av+1);
 
-    const q = new Query(this, 0,
+    const q = new Query(this, this.MODULE_user,
 			this.PL_Q_ALLOW_YIELD|this.PL_Q_CATCH_EXCEPTION,
 			"wasm_call_string_with_heartbeat/3", av,
 			(a) => this.toJSON(a+2));
@@ -706,11 +716,11 @@ class Prolog
 	{ if ( rc.yield !== undefined )
 	  { switch(rc.yield)
 	    { case "beat":
-	        return setTimeout(() => next_foreach(rc.resume("true")), 0);
+		return setTimeout(() => next_foreach(rc.resume("true")), 0);
 	      case "builtin":
-                return rc.resume((rc) => next_foreach(rc));
+		return rc.resume((rc) => next_foreach(rc));
 	      default:		// unsupported yield
-	        throw(rc);
+		throw(rc);
 	    }
 	  } else if ( rc.value )
 	  { if ( callback )
@@ -934,6 +944,11 @@ class Prolog
   { this.with_frame(() =>
     { const term = this.toProlog(obj, undefined, {string:"string"});
 
+      if ( !term )
+      { console.log("Could not convert", obj);
+	throw("Could not convert JavaScript data to Prolog");
+      }
+
       this.bindings.WASM_set_yield_result(term);
     }, true);
   }
@@ -1121,8 +1136,8 @@ class Prolog
 	  rc = prolog.bindings.PL_put_nil(term);
 
 	for(var i=data.length-1; i >= 0 && rc; i--)
-	{ toProlog(prolog, data[i], h, ctx);
-	  rc = prolog.bindings.PL_cons_list(term, h, term);
+	{ rc = ( toProlog(prolog, data[i], h, ctx) &&
+		 prolog.bindings.PL_cons_list(term, h, term) );
 	}
 
 	return rc;
@@ -1140,7 +1155,7 @@ class Prolog
 	  break;
 	case "string":
 	{ const flags = ctx.string === "string" ? prolog.PL_STRING
-		                                : prolog.PL_ATOM;
+						: prolog.PL_ATOM;
 	  rc = prolog.put_chars(term, data, flags);
 	  break;
 	}
@@ -1183,12 +1198,14 @@ class Prolog
 
 		if ( args !== undefined )
 		{ const av = prolog.new_term_ref(args.length);
-		  const f  = prolog.new_functor(prolog.new_atom(name), args.length);
+		  const f  = prolog.new_functor(prolog.new_atom(name),
+						args.length);
 
-		  for(var i=0; i<args.length; i++)
-		    toProlog(prolog, args[i], av+i, ctx);
+		  rc = true;
+		  for(var i=0; i<args.length && rc; i++)
+		    rc = toProlog(prolog, args[i], av+i, ctx);
 
-		  rc = prolog.bindings.PL_cons_functor_v(term, f, av);
+		  rc = rc && prolog.bindings.PL_cons_functor_v(term, f, av);
 		}
 		break;
 	      }
@@ -1208,8 +1225,13 @@ class Prolog
 	      }
 	      case "l":
 	      { rc = toList(term, data.v, data.t);
+		break;
+	      }
+	      default:
+	      { console.log(`Object with invalid $t:${data.$t}`);
 	      }
 	    }
+	    break;
 	  } else
 	  { switch( data.constructor.name )
 	    { case "ArrayBuffer":
@@ -1227,12 +1249,16 @@ class Prolog
 		if ( class_name != "Object" )
 		  tag = prolog.new_atom(class_name);
 
-		for(var i=0; i<len; i++)
-		{ toProlog(prolog, data[keys[i]], av+i, ctx);
-		  prolog.module.setValue(atoms+4*i, prolog.new_atom(keys[i]), 'i32');
+		rc = true;
+		for(var i=0; i<len && rc; i++)
+		{ rc = toProlog(prolog, data[keys[i]], av+i, ctx);
+		  prolog.module.setValue(atoms+4*i,
+					 prolog.new_atom(keys[i]),
+					 'i32');
 		}
 
-		rc = prolog.bindings.PL_put_dict(term, tag, len, atoms, av);
+		rc = rc && prolog.bindings.PL_put_dict(term, tag, len,
+						       atoms, av);
 		prolog.module._free(atoms);
 		break;
 	      }
@@ -1254,6 +1280,7 @@ class Prolog
 	    }
 	    break;
 	  }
+	  break;
 	default:
 	  return null;
       }
@@ -1279,9 +1306,9 @@ class Prolog
     flags  = flags||(this.CVT_ALL|this.CVT_WRITEQ);
     flags |= this.CVT_EXCEPTION|this.REP_UTF8;
     if (this.bindings.PL_get_chars(term, ptr, flags)) {
-        rc = this.module.UTF8ToString(this.module.getValue(ptr, 'i32'));
+	rc = this.module.UTF8ToString(this.module.getValue(ptr, 'i32'));
     } else {
-        rc = null;
+	rc = null;
     }
     this.module._free(ptr);
 
@@ -1364,16 +1391,20 @@ class Query {
     this.open   = true;
     this.argv   = argv;
     this.frame  = fid;
+    prolog.open_queries.push(this);
   }
 
   [Symbol.iterator]() { return this; }
 
   next()
-  { if ( !this.open )
+  { const prolog = this.prolog;
+    const argv   = this.argv;
+
+    if ( !this.open )
       return { done: true };
 
-    const prolog = this.prolog;
-    const argv   = this.argv;
+    if ( this != prolog.open_queries.at(-1) )
+      console.log("Attempt for Query.next() on not innermost query");
 
     switch(prolog.bindings.PL_next_solution(this.qid))
     { case prolog.PL_S_EXCEPTION:
@@ -1389,23 +1420,23 @@ class Query {
 	}
       }
       case prolog.PL_S_FALSE:
-        this.close();
+	this.close();
 	return { done: true };
       case prolog.PL_S_LAST:
 	this.close();
 	return { done: true,
-	         value: this.map ? this.map.call(this, argv) : argv
+		 value: this.map ? this.map.call(this, argv) : argv
 	       };
       case prolog.PL_S_TRUE:
 	return { done: false,
-	         value: this.map ? this.map.call(this, argv) : argv
+		 value: this.map ? this.map.call(this, argv) : argv
 	       };
       case prolog.PL_S_YIELD:
       { let request = prolog.yield_request();
 
 	return { done: false,
-	         value: null,
-	         yield: request,
+		 value: null,
+		 yield: request,
 		 resume: (value) =>
 		 { prolog.set_yield_result(value);
 		   return this.next();
@@ -1514,7 +1545,13 @@ class Query {
 
   close()
   { if ( this.open )
-    { this.prolog.bindings.PL_cut_query(this.qid);
+    { const prolog = this.prolog;
+
+      if ( this != prolog.open_queries.at(-1) )
+	console.log("Attempt for Query.close() on not innermost query");
+      prolog.open_queries.pop();
+
+      this.prolog.bindings.PL_cut_query(this.qid);
       if ( this.frame )
 	this.prolog.bindings.PL_discard_foreign_frame(this.frame);
       this.open = false;
@@ -1566,11 +1603,13 @@ function prolog_js_call(request, result)
 	  case "/": return args[0] / args[1];
 	  case "&": return args[0] & args[1];
 	  case "|": return args[0] | args[1];
+	  case "&&": return args[0] && args[1];
+	  case "||": return args[0] || args[1];
 	}
       }
 
       const func = obj[fname];
-      if ( func )
+      if ( typeof(func) === "function"  )
 	return func.apply(obj, args);
       else
 	console.log("ERROR: Function", fname, "is not defined on", obj);
@@ -1589,7 +1628,7 @@ function prolog_js_call(request, result)
 	      obj = window;
 	      break;
 	    default:
-	      obj = obj[next];
+	      obj = eval(next);
 	  }
 	} else
 	{ obj = obj[next];
@@ -1606,21 +1645,29 @@ function prolog_js_call(request, result)
     return obj;
   }
 
-  return prolog.with_frame(() =>
-  { const ar = prolog.toJSON(request, { string: "string" });
-    let obj;
+  try
+  { return prolog.with_frame(() =>
+    { const ar = prolog.toJSON(request, { string: "string" });
+      let obj;
 
-    if ( ar.setter )
-    { const target = eval_chain(ar.target);
-      const value  = eval_chain(ar.value);
-      target[ar.setter] = value;
-      obj = true;
-    } else
-    { obj = eval_chain(ar);
-    }
+      if ( ar.setter )
+      { const target = eval_chain(ar.target);
+	const value  = eval_chain(ar.value);
+	target[ar.setter] = value;
+	obj = true;
+      } else
+      { obj = eval_chain(ar);
+      }
 
-    return prolog.unify(result, prolog.toProlog(obj));
-  }, false);
+      return prolog.unify(result, prolog.toProlog(obj));
+    }, false);
+  } catch (e)
+  { return prolog.bindings.PL_raise_exception(
+      prolog.toProlog(new prolog.Compound("error",
+					  [ new prolog.Compound("js_error", [e.toString()]),
+					    new prolog.Var()
+					  ])));
+  }
 }
 
 
@@ -1644,35 +1691,13 @@ function release_registered_object(id)
 }
 
 
-if ( typeof window !== 'undefined' )
-{ window.js_add_script = function(text, opts)
-  { opts = opts||{};
-    let node;
-
-    if ( opts.id )
-    { if ( (node = document.getElementById(opts.id)) )
-      { node.textContent = text;
-      } else
-      { node = document.createElement("script");
-        node.id = opts.id;
-        node.textContent = text;
-        document.body.appendChild(node);
-      }
-    } else
-    { node = document.createElement("script");
-      node.textContent = text;
-      document.body.appendChild(node);
-    }
-  }
-}
-
 if ( BigInt.prototype.toJSON === undefined )
 { BigInt.prototype.toJSON = function ()
   { return this.toString();
   }
 }
 
-if ( typeof HTMLCollection === "object" )
+if ( HTMLCollection && HTMLCollection.prototype && !HTMLCollection.prototype.toList )
 { HTMLCollection.prototype.toList = function()
   { const ar = [];
 
@@ -1682,5 +1707,3 @@ if ( typeof HTMLCollection === "object" )
     return ar;
   }
 }
-
-
